from abc import ABC

import torch.nn as nn
import torch
import torch.nn.functional as F

from lib.models.ostrack.layers.frozen_bn import FrozenBatchNorm2d


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1,
         freeze_bn=False):
    if freeze_bn:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            FrozenBatchNorm2d(out_planes),
            nn.ReLU(inplace=True))
    else:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                      padding=padding, dilation=dilation, bias=True),
            nn.BatchNorm2d(out_planes),
            nn.ReLU(inplace=True))


def conv_gn(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1,
         freeze_bn=False):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
        nn.GroupNorm(32, out_planes),
        nn.ReLU(inplace=True))


class MLP(nn.Module):
    """ Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim]))

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x

class CenterPredictor(nn.Module, ABC):
    def __init__(self, inplanes=64, channel=256, feat_sz=20, stride=16, freeze_bn=False, up_sample=False):
        super(CenterPredictor, self).__init__()
        self.feat_sz = feat_sz
        self.stride = stride
        self.img_sz = self.feat_sz * self.stride
        self.up_sample = up_sample

        if self.up_sample:
            # corner predict
            self.conv1_ctr = conv(inplanes, channel, freeze_bn=freeze_bn)
            self.conv5_ctr = nn.Conv2d(channel, 1, kernel_size=1)

            # size regress
            self.conv1_offset = conv(inplanes, channel, freeze_bn=freeze_bn)
            self.conv5_offset = nn.Conv2d(channel, 2, kernel_size=1)

            # size regress
            self.conv1_size = conv(inplanes, channel, freeze_bn=freeze_bn)
            self.conv5_size = nn.Conv2d(channel, 2, kernel_size=1)
        else:
            # corner predict
            self.conv1_ctr = conv(inplanes, channel, freeze_bn=freeze_bn)
            self.conv2_ctr = conv(channel, channel // 2, freeze_bn=freeze_bn)
            self.conv3_ctr = conv(channel // 2, channel // 4, freeze_bn=freeze_bn)
            self.conv4_ctr = conv(channel // 4, channel // 8, freeze_bn=freeze_bn)
            self.conv5_ctr = nn.Conv2d(channel // 8, 1, kernel_size=1)

            # size regress
            self.conv1_offset = conv(inplanes, channel, freeze_bn=freeze_bn)
            self.conv2_offset = conv(channel, channel // 2, freeze_bn=freeze_bn)
            self.conv3_offset = conv(channel // 2, channel // 4, freeze_bn=freeze_bn)
            self.conv4_offset = conv(channel // 4, channel // 8, freeze_bn=freeze_bn)
            self.conv5_offset = nn.Conv2d(channel // 8, 2, kernel_size=1)

            # size regress
            self.conv1_size = conv(inplanes, channel, freeze_bn=freeze_bn)
            self.conv2_size = conv(channel, channel // 2, freeze_bn=freeze_bn)
            self.conv3_size = conv(channel // 2, channel // 4, freeze_bn=freeze_bn)
            self.conv4_size = conv(channel // 4, channel // 8, freeze_bn=freeze_bn)
            self.conv5_size = nn.Conv2d(channel // 8, 2, kernel_size=1)

        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

        if up_sample:
            self.conv5_ctr.bias.data.fill_(-2.19)

        # '''about coordinates and indexs'''
        # with torch.no_grad():
        #     self.indice = torch.arange(0, self.feat_sz).view(-1, 1) * self.stride
        #     # generate mesh-grid
        #     self.coord_x = self.indice.repeat((self.feat_sz, 1)) \
        #         .view((self.feat_sz * self.feat_sz,)).float().cuda()
        #     self.coord_y = self.indice.repeat((1, self.feat_sz)) \
        #         .view((self.feat_sz * self.feat_sz,)).float().cuda()

    def forward(self, x, gt_score_map=None):
        """ Forward pass with input x. """
        score_map_ctr, size_map, offset_map = self.get_score_map(x)

        # assert gt_score_map is None
        if gt_score_map is None:
            bbox = self.cal_bbox(score_map_ctr, size_map, offset_map)
        else:
            bbox = self.cal_bbox(gt_score_map.unsqueeze(1), size_map, offset_map)

        return score_map_ctr, bbox, size_map, offset_map

    def cal_bbox(self, score_map_ctr, size_map, offset_map, return_score=False):
        max_score, idx = torch.max(score_map_ctr.flatten(1), dim=1, keepdim=True)
        idx_y = idx // self.feat_sz
        # idx_y = torch.div(idx, self.feat_sz, rounding_mode='floor')
        idx_x = idx % self.feat_sz

        idx = idx.unsqueeze(1).expand(idx.shape[0], 2, 1)
        size = size_map.flatten(2).gather(dim=2, index=idx)
        offset = offset_map.flatten(2).gather(dim=2, index=idx).squeeze(-1)

        # bbox = torch.cat([idx_x - size[:, 0] / 2, idx_y - size[:, 1] / 2,
        #                   idx_x + size[:, 0] / 2, idx_y + size[:, 1] / 2], dim=1) / self.feat_sz
        # cx, cy, w, h
        bbox = torch.cat([(idx_x.to(torch.float) + offset[:, :1]) / self.feat_sz,
                          (idx_y.to(torch.float) + offset[:, 1:]) / self.feat_sz,
                          size.squeeze(-1)], dim=1)

        if return_score:
            return bbox, max_score
        return bbox

    def get_pred(self, score_map_ctr, size_map, offset_map):
        max_score, idx = torch.max(score_map_ctr.flatten(1), dim=1, keepdim=True)
        idx_y = idx // self.feat_sz
        idx_x = idx % self.feat_sz

        idx = idx.unsqueeze(1).expand(idx.shape[0], 2, 1)
        size = size_map.flatten(2).gather(dim=2, index=idx)
        offset = offset_map.flatten(2).gather(dim=2, index=idx).squeeze(-1)

        # bbox = torch.cat([idx_x - size[:, 0] / 2, idx_y - size[:, 1] / 2,
        #                   idx_x + size[:, 0] / 2, idx_y + size[:, 1] / 2], dim=1) / self.feat_sz
        return size * self.feat_sz, offset

    def get_score_map(self, x):

        def _sigmoid(x):
            y = torch.clamp(x.sigmoid_(), min=1e-4, max=1 - 1e-4)
            return y

        if self.up_sample:
            # ctr branch
            x_ctr1 = self.conv1_ctr(x)
            score_map_ctr = self.conv5_ctr(x_ctr1)

            # offset branch
            x_offset1 = self.conv1_offset(x)
            score_map_offset = self.conv5_offset(x_offset1)

            # size branch
            x_size1 = self.conv1_size(x)
            score_map_size = self.conv5_size(x_size1)
        else:
            # ctr branch
            x_ctr1 = self.conv1_ctr(x)
            x_ctr2 = self.conv2_ctr(x_ctr1)
            x_ctr3 = self.conv3_ctr(x_ctr2)
            x_ctr4 = self.conv4_ctr(x_ctr3)
            score_map_ctr = self.conv5_ctr(x_ctr4)

            # offset branch
            x_offset1 = self.conv1_offset(x)
            x_offset2 = self.conv2_offset(x_offset1)
            x_offset3 = self.conv3_offset(x_offset2)
            x_offset4 = self.conv4_offset(x_offset3)
            score_map_offset = self.conv5_offset(x_offset4)

            # size branch
            x_size1 = self.conv1_size(x)
            x_size2 = self.conv2_size(x_size1)
            x_size3 = self.conv3_size(x_size2)
            x_size4 = self.conv4_size(x_size3)
            score_map_size = self.conv5_size(x_size4)
        return _sigmoid(score_map_ctr), _sigmoid(score_map_size), score_map_offset


def build_box_head(cfg, hidden_dim=None):
    stride = cfg.MODEL.STRIDE
    # stride = 16
    if cfg.MODEL.HEAD_TYPE == "MLP" or cfg.MODEL.HEAD_TYPE == "RELATIVE":
        hidden_dim = cfg.MODEL.HIDDEN_DIM
        mlp_head = MLP(hidden_dim, hidden_dim, 4, 3)  # dim_in, dim_hidden, dim_out, 3 layers
        return mlp_head
    elif cfg.MODEL.HEAD_TYPE == "CENTER":
        in_channel = cfg.MODEL.HIDDEN_DIM
        if hidden_dim is not None:
            in_channel = hidden_dim
        out_channel = 256
        if cfg.MODEL.UP_SAMPLE:
            stride = 4
            in_channel = 64
            out_channel = 64
        feat_sz = int(cfg.DATA.SEARCH.SIZE / stride)
        center_head = CenterPredictor(inplanes=in_channel, channel=out_channel,
                                      feat_sz=feat_sz, stride=stride, up_sample=cfg.MODEL.UP_SAMPLE)
        return center_head
    else:
        raise ValueError("HEAD TYPE %s is not supported." % cfg.MODEL.HEAD_TYPE)
